from typing import Annotated, Optional

from pydantic import BaseModel, Field

from qualibrate_runner.core.models.common import RunError
from qualibrate_runner.core.models.enums import RunStatusEnum


class WorkflowStatus(BaseModel):
    """Model representing the status of a workflow execution."""

    active: Annotated[
        bool,
        Field(description="Indicates if the workflow is currently running."),
    ]
    status: Annotated[
        RunStatusEnum, Field(description="The current status of the workflow.")
    ]
    active_node_name: Annotated[
        Optional[str],
        Field(description="The name of the currently active node, if any."),
    ]
    nodes_completed: Annotated[
        int,
        Field(
            description=(
                "The number of nodes that have been completed in the workflow."
            ),
        ),
    ]
    nodes_total: Annotated[
        int, Field(description="The total number of nodes in the workflow.")
    ]
    run_duration: Annotated[
        float,
        Field(
            description=(
                "The total runtime duration of the workflow in seconds."
            ),
        ),
    ]
    error: Annotated[
        Optional[RunError],
        Field(
            description=(
                "Any error encountered during the workflow execution, "
                "if applicable."
            ),
        ),
    ] = None



import json
import logging
from collections.abc import Generator
from datetime import datetime
from functools import partial
from itertools import islice
from pathlib import Path
from sys import version_info
from typing import Any, Optional, TextIO, cast

from qualibrate.utils.logger_m import LazyInitLogger, logger
from qualibrate.utils.logger_utils.filters import filter_log_date
from qualibrate_config.models import QualibrateConfig

if version_info >= (3, 10):
    from types import NoneType  # type: ignore[attr-defined]
else:
    NoneType = type(None)

__all__ = [
    "parse_log_line",
    "parse_log_line_with_previous",
    "get_logs_from_qualibrate_files",
    "get_logs_from_qualibrate_in_memory_storage",
]

default_asctime_log_format = "%Y-%m-%d %H:%M:%S,%f"


def parse_log_line(
    line: str, previous_msg: Optional[dict[str, Any]] = None
) -> dict[str, Any]:
    # JSON format
    if line.startswith("{"):
        try:
            data = dict(json.loads(line))
        except json.decoder.JSONDecodeError as ex:
            logging.exception(f"Can't parse line {line}", exc_info=ex)
            return {}
        if "asctime" in data:
            data["asctime"] = datetime.strptime(
                data["asctime"], default_asctime_log_format
            )
        return data
    # old default-string-format
    parts = line.split(" - ", maxsplit=3)
    if len(parts) == 4:
        return {
            "asctime": datetime.strptime(parts[0], default_asctime_log_format),
            "name": parts[1],
            "levelname": parts[2],
            "message": parts[3].rstrip(),
        }
    if previous_msg is None:
        return {"message": line.rstrip()}
    copied = previous_msg.copy()
    copied["message"] = line.rstrip()
    return copied


def parse_log_line_with_previous(
    file: TextIO,
) -> Generator[dict[str, Any], None, NoneType]:
    previous = None
    for line in file:
        previous = parse_log_line(line, previous)
        try:
            yield previous
        except TypeError as e:
            logging.exception(line, exc_info=e)
    return None


def get_logs_from_qualibrate_files(
    after: Optional[datetime] = None,
    before: Optional[datetime] = None,
    num_entries: int = 100,
    *,
    config: QualibrateConfig,
) -> list[dict[str, Any]]:
    log_folder = config.log_folder
    if log_folder is None:
        return []
    out_logs: list[dict[str, Any]] = []
    q_log_files = filter(Path.is_file, log_folder.glob("qualibrate.log*"))
    filter_log_date_range = partial(filter_log_date, after=after, before=before)
    for log_file in sorted(q_log_files):
        with open(log_file) as f:
            filtered = list(
                filter(filter_log_date_range, parse_log_line_with_previous(f))
            )
            lines_date_filtered = reversed(filtered)
            file_logs = islice(lines_date_filtered, num_entries - len(out_logs))
            out_logs.extend(file_logs)
            if len(out_logs) == num_entries:
                return list(reversed(out_logs))
    return list(reversed(out_logs))


def get_logs_from_qualibrate_in_memory_storage(
    after: Optional[datetime] = None,
    before: Optional[datetime] = None,
    num_entries: int = 100,
    *,
    config: QualibrateConfig,
) -> list[dict[str, Any]]:
    return cast(LazyInitLogger, logger).in_memory_handler.get_logs(
        after, before, num_entries
    )



from collections.abc import Mapping
from datetime import datetime
from typing import Annotated, Any, Optional, Union

from pydantic import AwareDatetime, BaseModel, Field, computed_field
from qualibrate.models.run_summary.graph import GraphRunSummary
from qualibrate.models.run_summary.node import NodeRunSummary

from qualibrate_runner.core.models.common import RunError, StateUpdate
from qualibrate_runner.core.models.enums import RunnableType, RunStatusEnum

__all__ = ["LastRun"]


class LastRun(BaseModel):
    """Model representing the last executed run."""

    status: Annotated[
        RunStatusEnum,
        Field(
            description=(
                "The status of the run. "
                f"Possible options: {tuple(v.value for v in RunStatusEnum)}."
            ),
        ),
    ]
    started_at: Annotated[
        AwareDatetime, Field(..., description="The start time of the run.")
    ]
    completed_at: Annotated[
        Optional[AwareDatetime],
        Field(description="The completion time of the run."),
    ] = None
    name: Annotated[str, Field(description="The name of the run.")]
    idx: Annotated[int, Field(..., description="The index of the run.")]
    runnable_type: Annotated[
        RunnableType,
        Field(
            description=(
                "The type of the runnable entity."
                f"Possible options: {tuple(v.value for v in RunnableType)}."
            ),
        ),
    ]
    passed_parameters: Annotated[
        Mapping[str, Any],
        Field(
            default_factory=lambda: dict(),
            description="The parameters passed to the run.",
        ),
    ]
    run_result: Annotated[
        Optional[Union[NodeRunSummary, GraphRunSummary]],
        Field(
            description=(
                "The result of the run. Can be result of node or graph."
            ),
        ),
    ] = None
    # Here is not using Annotated because of mypy issue.
    # It doesn't understand default factory as default value so expect
    # argument on init
    state_updates: Mapping[str, StateUpdate] = Field(
        default_factory=dict,
        description="The state updates during the run.",
    )
    error: Annotated[
        Optional[RunError],
        Field(description="Any error encountered during the run."),
    ] = None

    @computed_field(description="Duration of the run in seconds.")
    def run_duration(self) -> float:
        duration = (
            self.completed_at - self.started_at
            if self.completed_at is not None
            else datetime.now().astimezone() - self.started_at
        )
        return round(duration.total_seconds(), 3)



from typing import Any

from qualibrate import (
    NodeParameters,
    QualibrationGraph,
    QualibrationLibrary,
    QualibrationNode,
)

__all__ = ["QNodeType", "QGraphType", "QLibraryType"]

QNodeType = QualibrationNode[NodeParameters, Any]
QGraphType = QualibrationGraph[QNodeType]
QLibraryType = QualibrationLibrary[QNodeType]



from typing import Any, Union

from pydantic import BaseModel, Field

__all__ = ["RunError", "StateUpdate"]


class RunError(BaseModel):
    """Model representing an error encountered during execution."""

    error_class: str = Field(..., description="The class of the error.")
    message: str = Field(..., description="The error message.")
    traceback: list[str] = Field(..., description="The traceback of the error.")


class StateUpdate(BaseModel):
    key: str
    attr: Union[str, int]
    old: Any
    new: Any
    updated: bool = False



from pydantic import BaseModel

next_line = "\n"


def get_model_docstring(model: type[BaseModel]) -> str:
    return f"""
**Model**: {model.__name__}

{model.__doc__}

**Fields**:
```
{next_line.join(
        f'{name}: {info.description}' 
        for name, info in model.model_fields.items()
    )}
```
"""



from collections import defaultdict
from collections.abc import Hashable
from typing import Any, Generic, TypeVar

from fastapi import WebSocket


class SocketConnectionManagerList:
    def __init__(self) -> None:
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket) -> None:
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket) -> None:
        if websocket in self.active_connections:
            self.active_connections.remove(websocket)

    async def broadcast(self, message: Any) -> None:
        for connection in self.active_connections:
            await connection.send_json(message)

    @property
    def any_subscriber(self) -> bool:
        return len(self.active_connections) > 0


KT = TypeVar("KT", bound=Hashable)


class SocketConnectionManagerMapping(Generic[KT]):
    def __init__(self) -> None:
        self.active_connections: dict[KT, list[WebSocket]] = defaultdict(list)

    async def connect(self, key: KT, websocket: WebSocket) -> None:
        await websocket.accept()
        self.active_connections[key].append(websocket)

    def disconnect(self, key: KT, websocket: WebSocket) -> None:
        self.active_connections[key].remove(websocket)

    async def broadcast(self, key: KT, message: Any) -> None:
        if not self.any_subscriber_for(key):
            return
        for connection in self.active_connections[key]:
            await connection.send_json(message)

    @property
    def any_subscriber(self) -> bool:
        return len(self.active_connections) > 0 and any(
            map(len, self.active_connections.values())
        )

    def any_subscriber_for(self, key: KT) -> bool:
        return (
            key in self.active_connections
            and len(self.active_connections[key]) > 0
        )



from typing import Optional, Union

from pydantic import BaseModel, ConfigDict

from qualibrate_runner.core.models.enums import RunStatusEnum
from qualibrate_runner.core.models.last_run import LastRun
from qualibrate_runner.core.types import QGraphType, QNodeType


class State(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    last_run: Optional[LastRun] = None
    run_item: Optional[Union[QNodeType, QGraphType]] = None

    @property
    def is_running(self) -> bool:
        return (
            self.last_run is not None
            and self.last_run.status == RunStatusEnum.RUNNING
        )



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from fastapi import FastAPI

from qualibrate_runner.api.sockets.tasks import execution_history, run_status

__all__ = ["app_lifespan"]


@asynccontextmanager
async def app_lifespan(app: FastAPI) -> AsyncIterator[None]:
    await asyncio.gather(run_status(), execution_history())
    yield



from functools import cache

from qualibrate_runner.core.app.ws_manager import (
    SocketConnectionManagerList,
    SocketConnectionManagerMapping,
)


@cache
def get_run_status_socket_manager() -> SocketConnectionManagerList:
    return SocketConnectionManagerList()


@cache
def get_execution_history_socket_manager() -> (
    SocketConnectionManagerMapping[bool]
):
    return SocketConnectionManagerMapping[bool]()
